import os
import boto3
from botocore.exceptions import ClientError
from dataclasses import dataclass
from abc import ABC, abstractmethod


class NotFoundError(Exception):
    """
    Raised if a config value is not found.
    """

    pass


class StringLoader(ABC):
    @abstractmethod
    def load_string(self, field_name: str) -> str:
        """
        Load the config string value.

        If the config value is not found, this method
        should return NotFoundError with a descriptive
        error message indicating where the loader looked
        for the value.
        """
        pass


@dataclass
class Secret:
    ref: str
    """a reference to where the secret can be found"""

    value: str
    """the actual contents of the secret"""


class SecretStringLoader(ABC):
    @abstractmethod
    def load_secret_string(self, field_name: str) -> Secret:
        """
        Load the config secret string value.

        If the config value is not found, this method
        should return NotFoundError with a descriptive
        error message indicating where the loader looked
        for the value.
        """
        pass


class EnvLoader(StringLoader):
    def load_string(self, field_name: str) -> str:
        """
        The field name is transformed to an env var in the following
        format - `PROVIDER_CONFIG_FIELD_NAME`

        For example, if the field is `api_url`, the env var will be
        `PROVIDER_CONFIG_API_URL`
        """
        env_var = f"PROVIDER_CONFIG_{field_name.upper()}"
        value = os.getenv(env_var)
        if value is None:
            raise NotFoundError(f"{env_var} environment variable is not set")

        return value


class DevEnvSecretLoader(SecretStringLoader):
    """
    For local development only. Not supported in any deployed provider runtimes.
    """

    def load_secret_string(self, field_name: str) -> Secret:
        """
        The field name is transformed to an env var in the following
        format - `PROVIDER_SECRET_FIELD_NAME`

        For example, if the field is `api_token`, the env var will be
        `PROVIDER_SECRET_API_TOKEN`
        """
        env_var = f"PROVIDER_SECRET_{field_name.upper()}"

        value = os.getenv(env_var)
        if value is None:
            raise NotFoundError(f"{env_var} environment variable is not set")

        return Secret(ref=f"env://{env_var}", value=value)


class DictLoader(StringLoader, SecretStringLoader):
    """
    Used only for writing unit tests which require loaded config.
    """

    def __init__(self, config_dict: dict) -> None:
        self.config = config_dict

    def load_string(self, field_name: str) -> str:
        value = self.config.get(field_name, None)
        if value is None:
            raise NotFoundError(f"{field_name} is not set in config_dict")

        return value

    def load_secret_string(self, field_name: str) -> Secret:
        val = self.config.get(field_name, None)
        if val is None:
            raise NotFoundError(f"secret {field_name} is not set in config_dict")

        return Secret(ref=f"dict://{field_name}", value=val)


class SSMSecretLoader(SecretStringLoader):
    """
    Secret Loader for AWS SSM
    """

    def load_secret_string(self, field_name: str) -> Secret:
        """
        The field name is transformed to an env var in the following
        format - `PROVIDER_SECRET_FIELD_NAME`

        For example, if the field is `api_token`, the env var will be
        `PROVIDER_SECRET_API_TOKEN`
        """
        env_var = f"PROVIDER_SECRET_{field_name.upper()}"

        secret_path = os.getenv(env_var)
        if secret_path is None:
            raise NotFoundError(f"{env_var} environment variable is not set")

        client = boto3.client("ssm")
        try:
            res = client.get_parameter(Name=secret_path, WithDecryption=True)

            return Secret(ref=f"{secret_path}", value=res["Parameter"]["Value"])

        except ClientError as e:
            if e.response["Error"]["Code"] == "ParameterNotFound":
                raise NotFoundError(
                    f"The AWS SSM parameter '{secret_path}' was not found"
                )
            else:
                raise e



import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider import diagnostics, namespace, provider, health


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.fixture
def provider_with_config_validators():
    class ExampleProvider(provider.Provider):
        pass

    @provider.config_validator(name="List Users")
    def can_list_users(provider: ExampleProvider) -> None:
        provider.diagnostics.info("some message here")

    @provider.config_validator(name="Fails")
    def fails(provider: ExampleProvider) -> None:
        raise Exception("something bad happened")

    return ExampleProvider


def test_provider_config_validation_works(
    provider_with_config_validators, snapshot_json
):
    ExampleProvider = provider_with_config_validators
    prov = ExampleProvider()
    health.validate_config(prov)
    actual = prov.diagnostics.export_logs()
    assert actual == snapshot_json

    healthy = prov.healthy()
    assert healthy == False



import importlib
import json
import os
import pkgutil
import sys
from commonfate_provider import config
from commonfate_provider.runtime.aws_lambda import AWSLambdaRuntime
from commonfate_provider.runtime.initialise import initialise_provider
from commonfate_provider.schema import export_schema
from contextlib import redirect_stdout
import click


def import_submodules(package, recursive=True):
    """Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


@click.command()
def schema():
    cwd = os.getcwd()

    dirname = os.path.basename(cwd)
    parent_folder = os.path.abspath(os.path.join(dirname, "..", ".."))

    sys.path.append(parent_folder)
    import_submodules(dirname)

    schema = export_schema()
    print(json.dumps(schema))


@click.command()
@click.argument("event")
def run(event):
    """
    Execute a provider.
    """
    cwd = os.getcwd()

    dirname = os.path.basename(cwd)
    parent_folder = os.path.abspath(os.path.join(dirname, "..", ".."))

    sys.path.append(parent_folder)
    import_submodules(dirname)

    provider = initialise_provider(configurer=config.DEV_LOADER)

    runtime = AWSLambdaRuntime(
        provider=provider,
    )
    event_json = json.loads(event)

    # redirect stdout to stderr, in case the provider logs any
    # messages using print()
    with redirect_stdout(sys.stderr):
        result = runtime.handle(event=event_json, context=None)
    print(json.dumps(result))


@click.group()
def cli():
    pass


cli.add_command(schema)
cli.add_command(run)

if __name__ == "__main__":
    cli()



import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider.runtime import AWSLambdaRuntime
from commonfate_provider import config, namespace, provider, access, target
from commonfate_provider.tests import helper


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


@pytest.fixture
def runtime_fixture():
    class BasicProvider(provider.Provider):
        pass

    @access.target(kind="Default")
    class Args:
        group = target.String()
        pass

    @access.grant(kind="Default")
    def grant(p: BasicProvider, subject, args):
        pass

    basic_provider = BasicProvider()

    runtime = AWSLambdaRuntime(provider=basic_provider)
    return runtime


def test_lambda_handler_works(runtime_fixture: AWSLambdaRuntime):
    event = {
        "type": "grant",
        "data": {
            "subject": "testuser",
            "target": {"arguments": {"group": "test"}, "kind": "Default"},
        },
    }
    runtime_fixture.handle(event=event, context=None)


def test_provider_describe(runtime_fixture: AWSLambdaRuntime, snapshot_json):
    event = {"type": "describe"}
    actual = runtime_fixture.handle(event=event, context=None)
    assert actual == snapshot_json


def test_provider_describe_with_config(snapshot_json):
    class Provider(provider.Provider):
        api_url = provider.String(usage="API URL")
        api_key = provider.String(usage="API key", secret=True)

    p = helper.initialise_test_provider(
        {"api_url": "https://example.com", "api_key": "abcdef"}
    )
    runtime = AWSLambdaRuntime(
        provider=p, publisher="acmecorp", name="test", version="v0.1.0"
    )

    event = {"type": "describe"}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_provider_describe_with_errors(snapshot_json):
    class Provider(provider.Provider):
        api_url = provider.String(usage="API URL")
        api_key = provider.String(usage="API key", secret=True)

    p = helper.initialise_test_provider(
        {"api_url": "https://example.com", "api_key": "abcdef"}
    )
    runtime = AWSLambdaRuntime(
        provider=p, publisher="acmecorp", name="test", version="v0.1.0"
    )
    p.diagnostics.error("some error happened!")

    event = {"type": "describe"}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json



import typing
from dataclasses import dataclass
from abc import ABC
from commonfate_provider import diagnostics, namespace


@dataclass
class Field:
    usage: typing.Optional[str] = None
    secret: bool = False
    optional: bool = False


class String(Field):
    def set(self, val: str) -> None:
        self._value = val

    def get(self) -> str:
        return self._value


class Provider(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._safe_config = {}
        """
        Used internally by the provider library
        to represent a 'safe' configuration with
        any secrets censored out and replaced
        with references.
        """

        self.diagnostics = diagnostics.Logs()
        """
        User-facing log messages associated with
        a provider.

        If any error messages are logged,
        the provider will be marked as unhealthy.
        """

    def __init_subclass__(cls) -> None:
        namespace.register_provider(cls)
        return super().__init_subclass__()

    def healthy(self) -> bool:
        """
        Built-in method to determine whether the provider is
        healthy and able to receive requests.
        """
        return self.diagnostics.has_no_errors()

    def setup(self):
        """
        Provider-specific setup logic can be called here, such as creating API clients.
        """
        pass

    @classmethod
    def export_schema(cls) -> dict:
        """
        Exports config variables defined in a Provider class
        to a dictionary.

        For example, if a Provider is defined as:
        ```
        class Provider(provider.Provider):
            url = provider.String()
        ```

        this method will produce a dictionary in the form:
        ```
        {"url": {"type": "string"}}
        ```
        """
        config_vars = {}
        all_vars = [(k, v) for (k, v) in vars(cls).items() if not k.startswith("__")]
        for k, v in all_vars:
            if type(v) == String:
                val: String = v

                config_vars[k] = {
                    "type": "string",
                    "usage": val.usage,
                    "secret": val.secret,
                }
        return config_vars


ConfigValidatorFunc = typing.Callable[[Provider], None]


@dataclass
class ConfigValidator:
    name: str
    func: ConfigValidatorFunc


def config_validator(
    name: str,
) -> typing.Callable[[ConfigValidatorFunc], ConfigValidatorFunc]:
    def actual_decorator(func: ConfigValidatorFunc):
        id = func.__name__
        cv = ConfigValidator(name=name, func=func)
        namespace.register_config_validator(id=id, config_validator=cv)
        return func

    return actual_decorator



from dataclasses import dataclass
from commonfate_provider import (
    provider,
    target,
    resources,
    tasks,
    namespace,
    schema,
)
import typing

from pydantic import BaseModel, Field


class GrantData(BaseModel):
    class Target(BaseModel):
        arguments: typing.Dict[str, str]
        kind: str

    subject: str
    target: Target


class Grant(BaseModel):
    type: typing.Literal["grant"]
    data: GrantData


class Revoke(BaseModel):
    type: typing.Literal["revoke"]
    data: GrantData


class Describe(BaseModel):
    type: typing.Literal["describe"]


class LoadResources(BaseModel):
    class Data(BaseModel):
        name: str
        """the resource loader function ID to run"""
        ctx: dict = {}
        """context information for the task"""

    type: typing.Literal["loadResources"]
    data: Data


class Event(BaseModel):
    __root__: typing.Union[Grant, Revoke, LoadResources, Describe] = Field(
        ..., discriminator="type"
    )


_T = typing.TypeVar("_T")


@dataclass
class AWSLambdaRuntime:
    provider: provider.Provider
    name: typing.Optional[str] = None
    version: typing.Optional[str] = None
    publisher: typing.Optional[str] = None

    def handle(self, event, context):
        parsed = Event.parse_obj(event)
        event = parsed.__root__

        if isinstance(event, Grant):
            try:
                registered_targets = namespace.get_target_classes()
                registered_target = registered_targets[event.data.target.kind]
                args_cls = registered_target.cls
            except KeyError:
                all_keys = ",".join(registered_targets.keys())
                raise KeyError(
                    f"unhandled target kind {event.data.target.kind}, supported kinds are [{all_keys}]"
                )

            args = target._initialise(args_cls, event.data.target.arguments)
            grant = registered_target.get_grant_func()
            grant(self.provider, event.data.subject, args)
            return {"message": "granting access"}

        elif isinstance(event, Revoke):
            try:
                registered_targets = namespace.get_target_classes()
                registered_target = registered_targets[event.data.target.kind]
                args_cls = registered_target.cls
            except KeyError:
                all_keys = ",".join(registered_targets.keys())
                raise KeyError(
                    f"unhandled target kind {event.data.target.kind}, supported kinds are [{all_keys}]"
                )

            args = target._initialise(args_cls, event.data.target.arguments)

            revoke = registered_target.get_revoke_func()
            revoke(self.provider, event.data.subject, args)
            return {"message": "revoking access"}
        if isinstance(event, Describe):
            # Describe returns the configuration of the provider including the current status.
            result = {}
            result["provider"] = {
                "publisher": self.publisher,
                "name": self.name,
                "version": self.version,
            }
            result["config"] = self.provider._safe_config
            result["diagnostics"] = self.provider.diagnostics.export_logs()
            result["healthy"] = self.provider.diagnostics.has_no_errors()
            result["schema"] = schema.export_schema()

            return {"body": result}

        elif isinstance(event, LoadResources):
            resources._reset()
            tasks._reset()
            tasks._execute(
                provider=self.provider, name=event.data.name, ctx=event.data.ctx
            )
            # find the resources and pending tasks, and return them
            found = resources.get()
            pending_tasks = tasks.get()
            print(found)
            response = {
                "resources": [r.export_json() for r in found],
                "pendingTasks": [t.json() for t in pending_tasks],
            }
            print(response)
            return {"body": response}

        else:
            raise Exception(f"unhandled event type")



import pytest
from commonfate_provider import resources, namespace


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


class ExampleResource(resources.Resource):
    value: str


def test_resource_equality():
    first = ExampleResource(id="test", value="test")
    second = ExampleResource(id="test", value="test")
    assert first.__eq__ is not None
    assert first == second



import pytest
from commonfate_provider import namespace, provider, config
from commonfate_provider.config import loaders


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


def test_configurer_works():
    class ExampleProvider(provider.Provider):
        value = provider.String()

    loader = loaders.DictLoader({"value": "something"})
    configurer = config.Configurer(string_loader=loader, secret_string_loader=loader)

    p = ExampleProvider()
    configurer.configure(p)

    assert p.value.get() == "something"


def test_configurer_works_with_secrets():
    class ExampleProvider(provider.Provider):
        value = provider.String(secret=True)

    loader = loaders.DictLoader({"value": "something"})
    secret_loader = loaders.DictLoader({"value": "something_secret"})
    configurer = config.Configurer(
        string_loader=loader, secret_string_loader=secret_loader
    )

    p = ExampleProvider()
    configurer.configure(p)

    # should be loaded through the secret loader
    assert p.value.get() == "something_secret"



import pytest
from commonfate_provider import namespace, target, access, resources
from commonfate_provider import provider


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


@access.target()
class ExampleArgs:
    group = target.String()


def test_parse_args():
    got_target = target._initialise(ExampleArgs, {"group": "test"})
    assert got_target.group == "test"


def test_parse_args_missing_required():
    with pytest.raises(target.ParseError):
        target._initialise(ExampleArgs, {})


def test_export_target_schema():
    @access.target()
    class ExampleTarget:
        my_property = target.String(title="MyProperty")

    got = target.export_schema()

    want = {
        "ExampleTarget": {
            "schema": {
                "my_property": {
                    "id": "my_property",
                    "resourceName": None,
                    "title": "MyProperty",
                    "type": "string",
                }
            }
        }
    }

    assert want == got


def test_export_target_resource_schema():
    class MyResource(resources.Resource):
        pass

    @access.target()
    class ExampleTarget:
        my_property = target.String(title="MyProperty")
        my_resource = target.Resource(title="MyResource", resource=MyResource)

    got = target.export_schema()

    want = {
        "ExampleTarget": {
            "schema": {
                "my_property": {
                    "id": "my_property",
                    "resourceName": None,
                    "title": "MyProperty",
                    "type": "string",
                },
                "my_resource": {
                    "id": "my_resource",
                    "resource": {
                        "properties": {"id": {"title": "Id", "type": "string"}},
                        "required": ["id"],
                        "type": "object",
                        "title": "MyResource",
                    },
                    "title": "MyResource",
                    "resourceName": "MyResource",
                    "type": "string",
                },
            }
        }
    }

    assert want == got



import pytest
from commonfate_provider import namespace, provider, diagnostics


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


@pytest.fixture
def example_provider():
    class ExampleProvider(provider.Provider):
        value = provider.String()

    return ExampleProvider


def test_export_schema_works(example_provider):
    ExampleProvider = example_provider
    got = ExampleProvider.export_schema()
    want = {"value": {"type": "string", "usage": None, "secret": False}}
    assert got == want



from commonfate_provider import namespace, resources, target


def export_schema() -> dict:
    """
    Export a Provider schema, ready to be serialized in JSON format.
    """

    schema = {}

    Provider = namespace.get_provider()

    schema["target"] = target.export_schema()
    schema["config"] = Provider.export_schema()
    schema["audit"] = resources.audit_schema()

    return schema



import typing

import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider import resources


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


class FirstResource(resources.Resource):
    """
    An example resource
    """

    value: str
    related_field: typing.Optional[str] = resources.Related("FirstResource")


class SecondResource(resources.Resource):
    value: str
    related_field: typing.Optional[str] = resources.Related(FirstResource)


@resources.fetcher
def load_resources():
    pass


def test_resource_schema_works(snapshot_json):
    got = resources.audit_schema()
    assert got == snapshot_json



import sys
from os import path

import pytest

from commonfate_provider import namespace


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


def test_entrypoint_works():
    sys.path.append(path.join(path.dirname(__file__)))
    from commonfate_provider.runtime import aws_lambda_entrypoint



import typing
from commonfate_provider import namespace, tasks
from pydantic import BaseModel, Field
import inspect


def composite_id(fields: typing.List[str]):
    return "/".join(fields)


class Resource(BaseModel):
    id: str

    def __init_subclass__(cls) -> None:
        namespace.register_resource_class(cls)
        return super().__init_subclass__()

    def export_json(self) -> dict:
        return {"type": self.__class__.__name__, "data": dict(self)}


T = typing.TypeVar("T", bound=Resource)


def Related(
    to: typing.Union[typing.Type[T], str], title: str = None, description: str = None
) -> str:
    if inspect.isclass(to):
        to = to.__name__
    return Field(relatedTo=to, title=title, description=description)


def fetcher(func: tasks.LoaderFunc):
    namespace.register_resource_loader(fetcher)
    return func


def audit_schema():
    """
    Returns the schema for the 'audit' section in a provider schema as a dict.
    This section defines the resources the provider can read, along with the
    fetching methods which can be called to fetch them.
    """
    loaders = {}
    for k in namespace.get_resource_loaders().keys():
        loaders[k] = {"title": k}

    resources = {}
    for Klass in namespace.get_resource_classes():
        properties = {}
        all_vars = [(k, v) for (k, v) in vars(Klass).items() if not k.startswith("__")]
        for k, v in all_vars:
            properties[k] = {"type": "string"}
        resources[Klass.__name__] = Klass.schema()

    return {"resourceLoaders": loaders, "resources": resources}


def register(resource: Resource):
    namespace._ALL_RESOURCES.append(resource)


def without_keys(d, keys):
    return {x: d[x] for x in d if x not in keys}



import typing
import boto3
from botocore.exceptions import ClientError
from dataclasses import dataclass
from commonfate_provider.config import loaders
from commonfate_provider import provider


class ConfigError(Exception):
    """
    Raised if the provider is incorrectly configured.
    """

    pass


@dataclass
class Configurer:
    """
    Configures a provider with the `configure()` method.
    """

    string_loader: loaders.StringLoader
    secret_string_loader: loaders.SecretStringLoader

    def configure(self, p: provider.Provider):
        """
        Configure a provider. This method looks at the defined fields in the provider class
        and tries to resolve them to actual values. Depending on whether the value
        is a secret or not, `string_loader` or `secret_string_loader` is used.
        These are defined when the Configurer is initialised.

        Example:
        ```
        class MyProvider(provider.Provider):
            api_url = provider.String()

        c = Configurer(string_loader=..., secret_string_loader=...)

        p = MyProvider()
        c.configure(p)

        # p.api_url will now be set, and can be used in the provider with
        # p.api_url.get()
        ```
        """
        all_vars = [
            (k, v) for (k, v) in vars(p.__class__).items() if not k.startswith("__")
        ]
        for k, v in all_vars:
            if isinstance(v, provider.String):
                self.resolve_string(p=p, key=k, val=v)

    def resolve_string(self, p: provider.Provider, key: str, val: provider.String):
        try:
            if val.secret is True:
                # the value is a secret
                resolved_secret = self.secret_string_loader.load_secret_string(key)
                val.set(resolved_secret.value)
                setattr(p, key, val)

                # set the value in the safe config dict to the ref,
                # as the value is sensitive.
                p._safe_config[key] = resolved_secret.ref
                return

            else:
                # the value is not a secret
                resolved_value = self.string_loader.load_string(key)
                val.set(resolved_value)
                setattr(p, key, val)

                # the value is not sensitive, so it can be shown to users.
                p._safe_config[key] = resolved_value
                return

        except loaders.NotFoundError as e:
            # if we get here, we didn't find a value for the config variable
            # this is fine if the variable is optional, but otherwise we raise an exception
            if val.optional is False:
                p.diagnostics.error(f"config {key} is required: {e}")


DEV_LOADER = Configurer(
    string_loader=loaders.EnvLoader(),
    secret_string_loader=loaders.DevEnvSecretLoader(),
)
"""
used only for local provider development.
"""

AWS_LAMBDA_LOADER = Configurer(
    string_loader=loaders.EnvLoader(),
    secret_string_loader=loaders.SSMSecretLoader(),
)
"""
used in the AWS Lambda runtime.
"""



from commonfate_provider import namespace, provider


def validate_config(p: provider.Provider):
    all_validators = namespace.get_config_validators()
    for id, validator in all_validators.items():
        try:
            validator.func(p)
        except Exception as e:
            p.diagnostics.error(f"config validation {id} failed: {e}")
